"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const defaults_1 = require("./defaults");
const helpers_1 = require("./helpers");
const utils_1 = require("./utils");
function matchCallbacks(node, position, ctx) {
    const runCallbacks = ctx.config.runCallbacks;
    let matched = [];
    if (runCallbacks) {
        if (node.isRoot && !ctx.config.rootObjectCallbacks) {
            return matched;
        }
        const callbacks = ctx.callbacksByPosition[position];
        if (typeof callbacks == 'undefined') {
            return [];
        }
        for (let i = 0; i < callbacks.length; ++i) {
            const callback = callbacks[i];
            // exit if nodeTypeFilters are defined and not in list
            if (typeof callback.nodeTypeFilters !== 'undefined' && callback.nodeTypeFilters.indexOf(node.nodeType) === -1) {
                continue;
            }
            // exit if keyFilters are defined and not in list
            if (typeof callback.keyFilters !== 'undefined'
                && (typeof node.keyInParent === 'number' || callback.keyFilters.indexOf(node.keyInParent) === -1)) {
                continue;
            }
            matched.push(callback);
        }
    }
    return matched;
}
function validateVisitation(val, ctx) {
    const mode = ctx.config.graphMode;
    let parseObject = true;
    if (ctx.seenObjects.indexOf(val) !== -1) {
        if (mode === 'finiteTree') {
            throw "The object violates the defined structure. Override 'graphMode' in the config to allow parsing different object structures.";
        }
        else if (mode === 'graph') {
            parseObject = false;
        } // otherwise, infinites are allowed
    }
    else {
        ctx.seenObjects.push(val);
    }
    return parseObject;
}
function execCallbacks(callbacks, node, ctx) {
    for (let i = 0; i < callbacks.length; ++i) {
        if (ctx.config.monitorPerformance) {
            const cbStackStart = new Date();
            callbacks[i].callback(node);
            node.executedCallbacks.push(callbacks[i]);
            // @ts-ignore
            ctx.report.callbackProcessingTime += (new Date() - cbStackStart);
        }
        else {
            callbacks[i].callback(node);
            node.executedCallbacks.push(callbacks[i]);
        }
    }
}
function getNormalizedType(val) {
    if (Array.isArray(val))
        return 'array';
    else if (typeof val === 'object')
        return 'object';
    return 'value';
}
function process(node, mode, ctx, queue) {
    let { val, path, nodeType } = node;
    const isNotValue = nodeType === 'object' || nodeType === 'array';
    if (isNotValue && !validateVisitation(val, ctx))
        return;
    const matchedPreCallbacks = matchCallbacks(node, 'preWalk', ctx);
    execCallbacks(matchedPreCallbacks, node, ctx);
    // add children to queue
    if (nodeType === 'array') {
        for (let i = 0; i < val.length; ++i) {
            const childData = {
                isArrayMember: true,
                keyInParent: i,
                rawType: typeof val[i],
                val: val[i],
                isRoot: false,
                path: path + ctx.config.pathFormat(i.toString(), true),
                parent: node,
                executedCallbacks: [],
                nodeType: getNormalizedType(val[i])
            };
            if (mode === 'breadth') {
                queue.push(childData);
            }
            else if (mode === 'depth') {
                process(childData, 'depth', ctx);
            }
        }
    }
    else if (nodeType === 'object') {
        for (const subKey in val) {
            if (val.hasOwnProperty(subKey)) {
                const childData = {
                    isArrayMember: false,
                    keyInParent: subKey,
                    val: val[subKey],
                    rawType: typeof val[subKey],
                    isRoot: false,
                    path: path + ctx.config.pathFormat(subKey, false),
                    parent: node,
                    executedCallbacks: [],
                    nodeType: getNormalizedType(val[subKey]),
                };
                if (mode === 'breadth') {
                    queue.push(childData);
                }
                else if (mode === 'depth') {
                    process(childData, 'depth', ctx);
                }
            }
        }
    }
    // match and run post-traverse callbacks
    const matchedPostCallbacks = matchCallbacks(node, 'postWalk', ctx);
    execCallbacks(matchedPostCallbacks, node, ctx);
}
function __depthTraverse(inputData, ctx) {
    process(inputData, 'depth', ctx, undefined);
}
function __breadthTraverse(inputData, ctx) {
    const queue = [];
    process(inputData, 'breadth', ctx, queue);
    while (queue.length > 0)
        process(queue.shift(), 'breadth', ctx, queue);
}
function buildContext(config) {
    const ctx = defaults_1.buildDefaultContext(config);
    ctx.config.callbacks.forEach(cb => {
        const callback = Object.assign(Object.assign({}, cb), { executionOrder: typeof cb.executionOrder == 'undefined' ? 0 : cb.executionOrder, positionFilters: typeof cb.positionFilters == 'undefined' || cb.positionFilters.length < 1 ? [defaults_1.defaultCallbackPosition] :
                cb.positionFilters });
        callback.positionFilters.forEach(position => {
            if (!(position in ctx.callbacksByPosition))
                ctx.callbacksByPosition[position] = [];
            ctx.callbacksByPosition[position].push(callback);
        });
    });
    for (const key in ctx.callbacksByPosition) {
        ctx.callbacksByPosition[key] = ctx.callbacksByPosition[key].sort(helpers_1.executionOrderSort);
    }
    return ctx;
}
const walk = (obj, config) => {
    const context = buildContext(config);
    const rootNode = ({
        keyInParent: undefined,
        val: obj,
        rawType: typeof obj,
        isArrayMember: false,
        isRoot: true,
        path: '',
        nodeType: 'object',
        parent: undefined,
        executedCallbacks: []
    });
    const fn = context.config.traversalMode === 'depth'
        ? __depthTraverse
        : __breadthTraverse;
    try {
        fn(rootNode, context);
    }
    catch (err) {
        if (!(err instanceof utils_1.Break))
            throw err;
    }
    return context;
};
exports.default = walk;
